'use strict'
// manages funds in user wallet and spends bitcoin

import { KeyPair } from "./KeyPair"
import { Address, Tx, VarInt } from 'bsv'
import { Forge } from 'txforge'
import { IndexClient } from "./IndexClient"
import { Envelopes } from "./Envelopes"
import { TokenEnvelope } from "./TokenEnvelope"

export class Wallet {
    protected _keypair: KeyPair
    protected _envelopes: Envelopes = new Envelopes()
    protected _indexclient: IndexClient

    protected _proofs: {[key:string]: any}
    protected _use_dsnt = false
    
    constructor(wif:string) {
        this._keypair = KeyPair.fromWif(wif)
        this._indexclient = new IndexClient()
        this._proofs = {}
    }
    
    get Address(): typeof Address {
        return this._keypair.toAddress()
    }
    get PublicKey(): string { return this._keypair.publicKey.toString() }
    get Balance():number { return this._envelopes.Balance }
    get Envelopes():Envelopes { return this._envelopes }
    get EnvelopesCopy():string { return JSON.parse(JSON.stringify(this._envelopes)) }
    get Unspents():any[] { return this._envelopes.utxos}
    set Unspents(val) { this._envelopes.utxos = val}
    set UseDsnt(val:boolean) { this._use_dsnt = val}
    get UnspentEnvelops():any[] { return this._envelopes.unspent}


    async sleep(delay = 350) {
        new Promise(r => setTimeout(r, delay));
    }

    // clear the wallet state, call utxofetch to get current utxos
    clear () {
        this._proofs={}
        this._envelopes.clear()
    }

    // gets utxos and totals the value of confirmed utxos, check for proof
    async utxofetch() {        
        const utxos = await this._indexclient.getUnspents(this.Address.toString())
    
        for (let i:number=0; i < utxos.length; i++) {
            let v = utxos[i].value 
            let h = utxos[i].height
            //TODO: look for existing
            const existing = this._proofs[utxos[i].tx_hash]
            if (!existing) {
                const envelope = new TokenEnvelope()
                envelope.utxo = utxos[i]
                const txns:any[] = []
                if(h > 0){
                    let proof = this._proofs[utxos[i].tx_hash]
                    if( !proof ) {
                        const txn = await this._indexclient.getTx(utxos[i].tx_hash)
                        const txns:any = []
                        proof = await this.gettxproof(txn, txns)
                        // let proof = await this.getTxProof(txn, txns)
                        this._proofs[utxos[i].tx_hash] = [proof]
                        envelope.txns=txns
                        // to prevent hitting 3/sec limit?
                        if(i > 0) await this.sleep()
                    }
                    envelope.proof = proof
                } else {
                    //TODO: if unconfirmed then walk the tx to find the proof(s)
                    const txn = await this._indexclient.getTx(utxos[i].tx_hash)
                    const txns:any = []
                    envelope.proof = await this.gettxproof(txn,txns)
                    envelope.txns=txns
                }
                if (envelope.utxo && !envelope.satoshis) envelope.satoshis = envelope.utxo.value
                // console.log(envelope)
                this._envelopes.add(envelope)
            }
        }
        return utxos
    }

    // generates a transaction to spend funds from wallet
    spend(to:string, amount:number, utxos?:any, param_dsnt?:boolean) {
        if (utxos === undefined || utxos.length === 0) {
            return null
        }
        //clone array objects
        const forgeutxo = JSON.parse(JSON.stringify(utxos))
        //translate WOC utxos to forge format
        if (forgeutxo) {
            for (let i=0; i< forgeutxo.length; i++) {
                const utxo = forgeutxo[i]
                if (utxo.tx_hash && utxo.value) {
                    forgeutxo[i] = {
                        txid:utxo.tx_hash,  // use the one from WOC, do not reverse
                        vout:utxo.tx_pos,     // UTXO output index (also accepts `outputIndex` or `txOutNum`)
                        satoshis:utxo.value, // UTXO amount (also accepts `amount`)
                        script:utxo.script    // Hex-encoded UTXO script
                    }
                }
            }
        }
        // console.log(`FORGE`, forgeutxo)
        const outputs:any = [{ to: to, satoshis: amount }]
        if (this._use_dsnt || param_dsnt) {
            //OP_FALSE OP_RETURN OP_PUSHDATA PROTOCOL_ID OP_PUSHDATA CALLBACK_MESSAGE
            //CTxOut(25, CScript([OP_FALSE, OP_RETURN, 0x746e7364, CallbackMessage(1, [LOCAL_HOST_IP], [0]).serialize()]))
            // 01012d38673c0100
            // version, 1 byte
            const version='01'
            // ip address count, 1 byte
            const ipcount='01'
            // list of addresses, count*address size
            const ips='2d38673c'
            // input count, 1 bytes
            const inputcount='01'
            // list of inputs, count*varint
            const inputs='0'
            const callback=`0x${version}${ipcount}${ips}${inputcount}${inputs}`
            outputs.push({ data: ['0x64736e74', callback] })
        }
        const forge = new Forge({
        inputs: [forgeutxo],
        outputs: outputs,
        changeTo: this._keypair.toAddress().toString()
        })

        const bsvkp = this._keypair.toBsvKeyPair()
        forge
        .build()
        .sign({ keyPair: bsvkp })
        return { 
            txid: forge.tx.id(),
            rawhex: forge.tx.toHex()
        }
    }

        // generates a transaction to spend envelopes
        spendEnvelopes(to:string, amount:number, envelopes?:any, param_dsnt?:boolean) {
            if (envelopes === undefined || envelopes.length === 0) {
                return null
            }
            //clone array objects
            const utxos = envelopes.utxos
            const forgeutxo = JSON.parse(JSON.stringify(utxos))
            //translate WOC utxos to forge format
            if (forgeutxo) {
                for (let i=0; i< forgeutxo.length; i++) {
                    const utxo = forgeutxo[i]
                    if (utxo.tx_hash && utxo.value) {
                        forgeutxo[i] = {
                            txid:utxo.tx_hash,  // use the one from WOC, do not reverse
                            vout:utxo.tx_pos,     // UTXO output index (also accepts `outputIndex` or `txOutNum`)
                            satoshis:utxo.value, // UTXO amount (also accepts `amount`)
                            script:utxo.script    // Hex-encoded UTXO script
                        }
                    }
                }
            }
            // console.log(`FORGE`, forgeutxo)
            const outputs:any = [{ to: to, satoshis: amount }]
            if (this._use_dsnt || param_dsnt) {
                //OP_FALSE OP_RETURN OP_PUSHDATA PROTOCOL_ID OP_PUSHDATA CALLBACK_MESSAGE
                //CTxOut(25, CScript([OP_FALSE, OP_RETURN, 0x746e7364, CallbackMessage(1, [LOCAL_HOST_IP], [0]).serialize()]))
                // 01012d38673c0100
                // version, 1 byte
                const version='01'
                // ip address count, 1 byte
                const ipcount='01'
                // list of addresses, count*address size
                const ips='2d38673c'
                // input count, 1 bytes
                const inputcount='01'
                // list of inputs, count*varint
                const inputs='0'
                const callback=`0x${version}${ipcount}${ips}${inputcount}${inputs}`
                outputs.push({ data: ['0x64736e74', callback] })
            }
            const forge = new Forge({
            inputs: [forgeutxo],
            outputs: outputs,
            changeTo: this._keypair.toAddress().toString()
            })
    
            const bsvkp = this._keypair.toBsvKeyPair()
            forge
            .build()
            .sign({ keyPair: bsvkp })
            return { 
                txid: forge.tx.id(),
                rawhex: forge.tx.toHex()
            }
        }
    
    // call this after transaction is confirmed spent
    // to update the wallet balance and utxos
    // DEPRECATE
    updateSpent(txraw:string, utxos:any) {
        this.Envelopes.markSpent(utxos)
        //add change back to our wallet so we can spend it
        if (txraw) {
            const tx = this.decode(txraw)
            // console.log(tx)
            for (let x =0; x<tx.txOuts.length; x++) {
                const txout = tx.txOuts[x]
                if (this.isChange(txout)) {
                    const envelope = new TokenEnvelope()
                    envelope.utxo = {
                        //txid:utxo.tx_hash,  // use the one from WOC, do not reverse
                        tx_hash: tx.id(),
                        //vout:utxo.tx_pos,     // UTXO output index (also accepts `outputIndex` or `txOutNum`)
                        tx_pos: x,
                        //satoshis:utxo.value, // UTXO amount (also accepts `amount`)
                        value: txout.valueBn.toNumber(),
                        //script:utxo.script    // Hex-encoded UTXO script
                        script: txout.script.toHex()
                    }
                    // we know its chained, server should have proof already
                    // envelope.proof = x
                    //make this full txn? careful with storing different formats
                    envelope.txns = [tx]
                    this.Envelopes.add(envelope)
                }
            }
        }
    }

    updateSpentEnvelopes(txraw:string, envelopes:any) {
        this.Envelopes.markSpentEnvelopes(envelopes)
        //add change back to our wallet so we can spend it
        if (txraw) {
            const tx = this.decode(txraw)
            for (let x =0; x<tx.txOuts.length; x++) {
                const txout = tx.txOuts[x]
                if (this.isChange(txout)) {
                    const envelope = new TokenEnvelope()
                    envelope.utxo = {
                        //txid:utxo.tx_hash,  // use the one from WOC, do not reverse
                        tx_hash: tx.id(),
                        //vout:utxo.tx_pos,     // UTXO output index (also accepts `outputIndex` or `txOutNum`)
                        tx_pos: x,
                        //satoshis:utxo.value, // UTXO amount (also accepts `amount`)
                        value: txout.valueBn.toNumber(),
                        //script:utxo.script    // Hex-encoded UTXO script
                        script: txout.script.toHex()
                    }
                    // we know its chained, server should have proof already
                    // envelope.proof = x
                    //make this full txn? careful with storing different formats
                    envelope.txns = [tx]
                    this.Envelopes.add(envelope)
                }
            }
        }
    }

    isChange(txout:any):boolean {
        // console.log(`isChange?`,txout.script.toHex(),txout.script.toString())
        // console.log(`ischange2?`, this._keypair.toScript().toHex())
        return txout.script.toHex() === this._keypair.toScript().toHex()
    }

    // decode raw tx string of bytes into a json bsv lib Tx
    decode(rawtx:string) {
        try {
            return Tx.fromBuffer(Buffer.from(rawtx,'hex'))
        }
        catch { return null }
    }

    gettxproof = async (txn:any, txns:any):Promise<any> => {
        txns.push(txn)
        // console.log(`TXN`,txn)
        //if (txn.blockheight) { console.log(`TX CONFIRMED in ${txn.blockheight}`) }
        //console.log(`IN COUNT`, txn.vin.length)
        for (let x=0; x < txn.vin.length; x++) {
            const input = txn.vin[x]
            //console.log(`${input.txid}:${input.vout}`)
            const txinput = await this._indexclient.getTx(input.txid)
            //console.log(`INPUT`,txinput.txid)
            txns.push(txinput)
            if (txinput.blockheight) { 
                //console.log(`PARENT INPUT CONFIRMED in ${txinput.blockheight}`)
            }
            if (txinput.blockheight) {
                const proof = await this._indexclient.getProof(txinput.txid)
                //console.log(`PROOF`, proof[0].merkleRoot)
                return proof
            } else {
                //console.log(`CHAINED`,txinput.txid)
                return await this.gettxproof(txinput,txns)
            }
        }
    }

    // //TODO: needs work, does not return anything
    // async getTxProof (txn:any, txns:any):Promise<any> {
    //     txns.push(txn)
    //     // console.log(`TXN`,txn)
    //     //if (txn.blockheight) { console.log(`TX CONFIRMED in ${txn.blockheight}`) }
    //     //console.log(txn)
    //     //console.log(`IN COUNT`, txn.vin.length)
    //     for (let x=0; x < txn.vin.length; x++) {
    //         const input = txn.vin[x]
    //         //console.log(`${input.txid}:${input.vout}`)
    //         const txinput = await this._indexclient.getTx(input.txid)
    //         //console.log(`INPUT`,txinput.txid)
    //         txns.push(txinput)
    //         if (txinput.blockheight) { 
    //             //console.log(`PARENT INPUT CONFIRMED in ${txinput.blockheight}`)
    //         }
    //         if (txinput.blockheight) {
    //             const proof = await this._indexclient.getProof(txinput.txid)
    //             //console.log(`PROOF`, proof[0].merkleRoot)
    //         } else {
    //             console.log(`CHAINED`,txinput.txid)
    //             await this.getTxProof(txinput,txns)
    //         }
    //     }
    // }

    // //walk up parents to get merkle proofs of inputs
    // //TODO: incomplete, GET IMPROVED CODE FROM proof.test.ts
    // async getProofs(txid:string):Promise<any> {
    //     const txn = await this._indexclient.getTx(txid)
    //     // console.log(`TXN`,txn)
    //     if (txn.blockheight) { console.log(`TX CONFIRMED in ${txn.blockheight}`) }
    //     txn.vin.forEach(async (input:any) => {
    //         console.log(`${input.txid}:${input.vout}`)
    //         const txinput = await this._indexclient.getTx(input.txid)
    //         console.log(`INPUT`,txinput.txid)
    //         if (txinput.blockheight) { console.log(`PARENT INPUT CONFIRMED in ${txinput.blockheight}`) }
    //         if (txinput.blockheight) {
    //             const proof = await this._indexclient.getProof(txinput.txid)
    //             console.log(`PROOF`, proof)
    //             if (proof) return proof
    //             return this.getProofs(txid)
    //         }
    //     })
    // }

}